@@ -69,17 +69,8 @@ from PySide.QtUiTools import QUiLoader
 import undoCommands
 import ogAST
 import ogParser
-
-if hasattr(sys, 'frozen'):
-    # Detect py2exe distribution (that does not support the __file__ construct)
-    try:
-        CUR_DIR = os.path.dirname(unicode
-            (sys.executable, sys.getfilesystemencoding()))
-    except TypeError:
-        # On Linux with pyinstaller the above will not work
-        CUR_DIR = os.path.dirname(os.path.realpath(__file__))
-else:
-    CUR_DIR = os.path.dirname(os.path.realpath(__file__))
+from Connectors import Connection, VerticalConnection, CommentConnection, \
+                       RakeConnection, JoinConnection
 
 LOG = logging.getLogger(__name__)
 
@@ -286,6 +277,9 @@ class EditableText(QGraphicsTextItem, object):
             Slot connected to the autocompletion popup,
             invoked when selection is made
         '''
+        if not(self.textInteractionFlags() & Qt.TextEditable):
+            self.completer.hide()
+            return
         text_cursor = self.textCursor()
         # Go back to the previously saved cursor position
         text_cursor.setPosition(self.cursor_position)
@@ -372,7 +366,7 @@ class EditableText(QGraphicsTextItem, object):
                 text_cursor.clearSelection()
                 self.setTextCursor(text_cursor)
             # If something has changed, check syntax and create undo command
-            if(self.oldSize != self.parentItem().boundingRect() or 
+            if(self.oldSize != self.parentItem().boundingRect() or
                                                     self.oldText != str(self)):
                 # Call syntax checker from item containing the text (if any)
                 self.parentItem().check_syntax()
@@ -423,108 +417,6 @@ class EditableText(QGraphicsTextItem, object):
             return ''
 
 
-# pylint: disable=R0904
-class Connection(QGraphicsPathItem, object):
-    ''' Connection between two symbols (arrow) '''
-    # Default connections are not selectable
-    default_cursor = Qt.ArrowCursor
-    hasParent = False
-
-    def __init__(self, parent, child, connectionPoint=False):
-        super(Connection, self).__init__(parent)
-        self.parent = parent
-        self.child = child
-        self.child_is_a_connection_point = connectionPoint
-        self.start_point = QPointF(0, 0)
-        self.end_point = QPointF(0, 0)
-        pen = QPen()
-        pen.setColor(Qt.blue)
-        pen.setCosmetic(False)
-        self.setPen(pen)
-        self.parent_rect = parent.sceneBoundingRect()
-        self.childRect = child.sceneBoundingRect()
-        # Activate cache mode to boost rendering by calling paint less often
-        self.setCacheMode(QGraphicsItem.DeviceCoordinateCache)
-
-    def __str__(self):
-        ''' Print connection information for debug purpose'''
-        return 'Connection: parent = {p}, child = {c}'.format(
-                p=str(self.parentItem()), c=str(self.child))
-
-    def reshape(self):
-        ''' Update the connection or arrow shape '''
-        new_shape = QPainterPath()
-        if(self.parentItem().terminal_symbol
-                and self.child_is_a_connection_point):
-            self.setPath(new_shape)
-            return
-        parent_rect = self.parentItem().boundingRect()
-        # Define connection start point
-        if hasattr(self.parentItem(), 'connectionPoint') and not isinstance(
-                self.child, (HorizontalSymbol, Comment)):
-            self.start_point = self.parentItem().connectionPoint
-        elif isinstance(self.child, Comment):
-            self.start_point = QPointF(
-                    parent_rect.width(), parent_rect.height() / 2)
-        else:
-            self.start_point = QPointF(
-                    parent_rect.width() / 2, parent_rect.height())
-        # Defined connection end point
-        if self.child_is_a_connection_point:
-            connection_point_scene = self.child.mapToScene(
-                    self.child.connectionPoint)
-            connection_point_local = self.mapFromScene(
-                    connection_point_scene)
-            self.end_point = connection_point_local
-        elif isinstance(self.child, Comment):
-            if self.child.on_the_right:
-                self.end_point = QPointF(self.child.x(),
-                        self.child.y() +
-                        self.child.boundingRect().height() / 2)
-            else:
-                self.end_point = QPointF(self.child.x() +
-                        self.child.boundingRect().width(),
-                        self.child.y() +
-                        self.child.boundingRect().height() / 2)
-        else:
-            self.end_point = QPointF()
-            self.end_point.setY(self.child.pos().y())
-        # Move to start point and draw the connection
-        new_shape.moveTo(self.start_point)
-        if not self.child_is_a_connection_point and not isinstance(
-                self.child, Comment):
-            if isinstance(self.child, HorizontalSymbol):
-                self.end_point.setX(self.child.pos().x() +
-                        self.child.boundingRect().width() / 2)
-                new_shape.lineTo(self.start_point.x(),
-                        self.start_point.y() + 10)
-                new_shape.lineTo(self.end_point.x(), self.start_point.y() + 10)
-            else:
-                self.end_point.setX(self.start_point.x())
-        elif isinstance(self.child, Comment):
-            # Make sure the connection does not overlap the comment item
-            if (self.child.on_the_right or
-                    (not self.child.on_the_right and
-                        self.child.x() + self.child.boundingRect().width()
-                        < self.parentItem().boundingRect().width())):
-                go_to_point = self.start_point.x() + 5
-            else:
-                go_to_point = self.end_point.x() + 5
-            new_shape.lineTo(go_to_point, self.start_point.y())
-            new_shape.lineTo(go_to_point, self.end_point.y())
-            new_shape.lineTo(self.end_point.x(), self.end_point.y())
-        else:
-            new_shape.lineTo(self.start_point.x(), self.end_point.y() - 10)
-            new_shape.lineTo(self.end_point.x(), self.end_point.y() - 10)
-        new_shape.lineTo(self.end_point)
-        # If required draw an arrow head (e.g. in SDL NEXTSTATE and JOIN)
-        if self.child.arrow_head:
-            new_shape.lineTo(self.end_point.x() - 5, self.end_point.y() - 5)
-            new_shape.moveTo(self.end_point)
-            new_shape.lineTo(self.end_point.x() + 5, self.end_point.y() - 5)
-        self.setPath(new_shape)
-
-
 # pylint: disable=R0904, R0902
 class Symbol(QObject, QGraphicsPathItem, object):
     '''
@@ -537,6 +429,8 @@ class Symbol(QObject, QGraphicsPathItem, object):
     _unique_followers = []  # unique : e.g. comment symbol
     _insertable_followers = []  # no limit to insert below current symbol
     _terminal_followers = []  # cannot be inserted between two symbols
+    # By default a symbol is resizeable
+    resizeable = True
     # By default symbol size may expand when inner text exceeds border
     auto_expand = True
     # By default connections between symbols are lines, not arrows
@@ -555,12 +449,15 @@ class Symbol(QObject, QGraphicsPathItem, object):
     needs_parent = True
     # nested_scene : a symbol may have content that can be visible on demand
     # (e.g. a subscene that appears when double-clicking on the item)
-    nested_scene = None
+    _allow_nesting = False
+    _nested_scene = None
     # keywords for the syntax highlighter
     blackbold = ()
     redbold = ()
     # Specify if the symbol can be drawn with anti-aliasing
     _antialiasing = True
+    # Specify if the symbol has a text area
+    has_text_area = True
 
     def __init__(self, parent=None):
         '''
@@ -613,6 +510,13 @@ class Symbol(QObject, QGraphicsPathItem, object):
 
     position = Property(QPointF, _pos, _set_pos)
 
+    def is_composite(self):
+        ''' Return True if nested scene has something in it '''
+        try:
+            return any(self._nested_scene.visible_symb)
+        except AttributeError:
+            return False
+
     @property
     def allowed_followers(self):
         '''
@@ -630,6 +534,23 @@ class Symbol(QObject, QGraphicsPathItem, object):
         #LOG.debug(str(followers))
         return followers
 
+    @property
+    def allow_nesting(self):
+        ''' Dynamically check if the scene can have a nested subscene
+            May be redefined in subclasses and checked on double-click
+        '''
+        return self._allow_nesting
+
+    @property
+    def nested_scene(self):
+        ''' Return the nested scene. Can be redefined '''
+        return self._nested_scene
+
+    @nested_scene.setter
+    def nested_scene(self, value):
+        ''' Set the value of the nested scene '''
+        self._nested_scene = value
+
     def closest_connection_point(self, coord):
         '''
            Given a position (QPointF), expected in this symbol's
@@ -712,7 +633,7 @@ class Symbol(QObject, QGraphicsPathItem, object):
             _, syntax_errors, ___, ____, _____ = (
                                 self.parser.parseSingleElement(
                                                 self.common_name, repr(self)))
-        except AssertionError:
+        except (AssertionError, AttributeError) as err:
             LOG.error('Checker failed - no parser for this construct?')
         else:
             try:
@@ -758,15 +679,14 @@ class Symbol(QObject, QGraphicsPathItem, object):
                 # to the connection point
                 for child in self.childItems():
                     # Find the connection point below
-                    if(isinstance(child, Connection) and
-                            child.child_is_a_connection_point):
+                    if isinstance(child, JoinConnection):
                         connection_below = child
                         break
                 if bep is not self:
                     # Item is not the branch entry point itself
                     bep.last_branch_item = self.parentItem()
                     connection_below.setParentItem(self.parentItem())
-                    self.parentItem().connectionBelow = connection_below
+                    #self.parentItem().connectionBelow = connection_below
                 else:
                     # delete the link to the connection point
                     connection_below.setParentItem(None)
@@ -797,7 +717,7 @@ class Symbol(QObject, QGraphicsPathItem, object):
             self.parentItem().cam(self.parentItem().pos(),
                                   self.parentItem().pos())
             self.setParentItem(None)
-        # Following line cause segfault on exit:
+        # Following line causes segfault on exit:
         #scene().removeItem(self)
         try:
             entrypoint_parent.updateConnectionPointPosition()
@@ -813,7 +733,7 @@ class Symbol(QObject, QGraphicsPathItem, object):
         ''' Return the connection above the symbol '''
         try:
             connection, = [cnx for cnx in self.parent.connections()
-                    if cnx.child == self]
+                           if cnx.child == self]
             return connection
         except ValueError:
             return None
@@ -883,6 +803,8 @@ class Symbol(QObject, QGraphicsPathItem, object):
 
     def resize_item(self, rect):
         ''' resize item, e.g. when editing text - move children accordingly '''
+        if not self.resizeable:
+            return
         pos = self.pos()
         delta_x = (self.boundingRect().width() - rect.width()) / 2.0
         delta_y = self.boundingRect().height() - rect.height()
@@ -890,7 +812,7 @@ class Symbol(QObject, QGraphicsPathItem, object):
         # Align children properly when resizing
         try:
             self.text.set_textbox_position()
-        except AttributeError:
+        except AttributeError as err:
             # if called before text is initialized - or if no textbox
             pass
         for child in self.childSymbols():
@@ -944,6 +866,10 @@ class Symbol(QObject, QGraphicsPathItem, object):
         else:
             self.mode = 'Move'
 
+    def double_click(self):
+        ''' Handle double click on symbol - redefined at symbol level '''
+        pass
+
     def mouse_move(self, event):
         ''' Handle resizing of items - moving is handled in subclass '''
         self.updateConnectionPoints()
@@ -1108,7 +1034,7 @@ class Symbol(QObject, QGraphicsPathItem, object):
             self.setPos(old_pos)
 
 
-class Comment(Symbol, object):
+class Comment(Symbol):
     '''
         Class used to handle right connected comments
     '''
@@ -1124,8 +1050,9 @@ class Comment(Symbol, object):
             local_pos = parent.mapFromScene(ast.pos_x, ast.pos_y)
             self.insert_symbol(parent, local_pos.x(), local_pos.y())
         self.set_shape(ast.width, ast.height)
-        self.text = EditableText(parent=self, text=ast.inputString,
-                hyperlink=ast.hyperlink)
+        self.text = EditableText(parent=self,
+                                 text=ast.inputString,
+                                 hyperlink=ast.hyperlink)
         self.common_name = 'end'
         self.parser = ogParser
 
@@ -1155,6 +1082,10 @@ class Comment(Symbol, object):
         self.connection = self.connect_to_parent()
         parent.cam(parent.pos(), parent.pos())
 
+    def connect_to_parent(self):
+        ''' Redefinition of the function to use a comment connector '''
+        return CommentConnection(self.parent, self)
+
     def delete_symbol(self):
         '''
             Specific delete actions for Comment:
@@ -1168,7 +1099,7 @@ class Comment(Symbol, object):
             Redefinition of the Resize function
             (Comment symbol only resizes in one direction)
         '''
-        if self.grabber.resize_mode.endswith('left'):
+        if not self.resizeable or self.grabber.resize_mode.endswith('left'):
             return
         self.set_shape(rect.width(), rect.height())
         self.update_connections()
@@ -1296,6 +1227,9 @@ class Cornergrabber(QGraphicsPolygonItem, object):
             # Parent item may have changed its cursor (e.g. when inserting
             # items). In that case, don't override the cursor for that area
             cursor = self.parent.cursor()
+        elif not self.parent.resizeable:
+            cursor = self.parent.default_cursor
+            self.resize_mode = ''
         # Left side
         elif 0.0 <= pos.x() <= 10.0:
             # Top-left corner: disabled
@@ -1309,7 +1243,6 @@ class Cornergrabber(QGraphicsPolygonItem, object):
                 self.resize_mode = 'left'
         # Middle of item
         elif 5.0 < pos.x() < self.boundingRect().width() - 10.0 and (
-                     # pos.y() <= 10.0 or
                      pos.y() > self.boundingRect().height() - 10.0):
             cursor = Qt.SizeVerCursor
             self.resize_mode = 'bottom'
@@ -1342,7 +1275,7 @@ class HorizontalSymbol(Symbol, object):
         super(HorizontalSymbol, self).__init__(parent)
         self.minDistanceToSymbolAbove = 20
         self.connection = None
-        if text:
+        if self.has_text_area:
             self.text = EditableText(parent=self, text=text,
                     hyperlink=hyperlink)
         if parent:
@@ -1351,6 +1284,10 @@ class HorizontalSymbol(Symbol, object):
         else:
             self.setPos(x or 0, y or 0)
 
+    def connect_to_parent(self):
+        ''' Redefined: connect to parent item '''
+        return RakeConnection(self.parent, self)
+
     def insert_symbol(self, parent, pos_x, pos_y):
         ''' Insert the symbol in the scene - Align below the parent '''
         if not parent:
@@ -1410,7 +1347,8 @@ class HorizontalSymbol(Symbol, object):
         ''' Return all the items's sibling symbols '''
         try:
             return (item for item in self.parent.childItems()
-                    if item is not self and type(item) is type(self))
+                    if item is not self and (isinstance(self, type(item)) or
+                        isinstance(item, type(self)))) # is type(self))
         except:
             return ()
 
@@ -1506,7 +1444,10 @@ class VerticalSymbol(Symbol, object):
             x=None, y=None, hyperlink=None):
         super(VerticalSymbol, self).__init__(parent)
         self.connection = None
-        self.text = EditableText(parent=self, text=text, hyperlink=hyperlink)
+        if self.has_text_area:
+            self.text = EditableText(parent=self,
+                                     text=text,
+                                     hyperlink=hyperlink)
         self.minDistanceToSymbolAbove = 15
         if parent:
             local_pos = self.mapFromScene(0, y or 0)
@@ -1521,6 +1462,10 @@ class VerticalSymbol(Symbol, object):
                 return symbol
         return None
 
+    def connect_to_parent(self):
+        ''' Redefined: connect to parent item with a straight line '''
+        return VerticalConnection(self.parent, self)
+
     def insert_symbol(self, parent, x, y):
         '''
             Vertical symbol: place the symbol in the scene.
