@@ -19,20 +19,24 @@
 
 __all__ = ['Input', 'Output', 'State', 'Task', 'ProcedureCall', 'Label',
            'Decision', 'DecisionAnswer', 'Join', 'Start', 'TextSymbol',
-           'Procedure', 'ProcedureStart', 'ProcedureStop', 'ASN1Viewer']
+           'Procedure', 'ProcedureStart', 'ProcedureStop', 'ASN1Viewer',
+           'StateStart', 'Process']
 
-from genericSymbols import(
-        HorizontalSymbol, VerticalSymbol, Connection, Comment)
+import traceback
+import logging
+from itertools import chain
+from collections import deque
 
 from PySide.QtCore import Qt, QPoint, QRect, QRectF
 from PySide.QtGui import(QPainterPath, QBrush, QColor, QPolygon,
         QRadialGradient, QPainter, QPolygonF, QPen)
 
+from genericSymbols import HorizontalSymbol, VerticalSymbol, Comment
+from Connectors import Connection, JoinConnection
+
 import ogParser
 import ogAST
 
-import traceback
-import logging
 
 LOG = logging.getLogger('sdlSymbols')
 
@@ -44,7 +48,7 @@ SDL_BLACKBOLD = ['\\b{word}\\b'.format(word=word) for word in (
                 'EXPONENT', 'TRUE', 'FALSE', 'MOD', 'FI', 'WRITE', 'WRITELN',
                 'LENGTH', 'PRESENT', 'FPAR', 'TODO', 'FIXME', 'XXX',
                 'CHECKME', 'PROCEDURE', 'EXTERNAL', 'IN', 'OUT', 'TIMER',
-                'SET_TIMER', 'RESET_TIMER')]
+                'SET_TIMER', 'RESET_TIMER', 'VIA', 'ENTRY', 'EXIT')]
 
 SDL_REDBOLD = ['\\b{word}\\b'.format(word=word) for word in (
               'INPUT', 'OUTPUT', 'STATE', 'DECISION', 'NEXTSTATE',
@@ -52,11 +56,11 @@ SDL_REDBOLD = ['\\b{word}\\b'.format(word=word) for word in (
 
 
 # pylint: disable=R0904
-class Input(HorizontalSymbol, object):
+class Input(HorizontalSymbol):
     ''' SDL INPUT Symbol '''
     _unique_followers = ['Comment']
     _insertable_followers = ['Task', 'ProcedureCall', 'Output', 'Decision',
-                        'Input',  'Label']
+                        'Input',  'Label', 'Connect']
     _terminal_followers = ['Join', 'State', 'ProcedureStop']
     completion_list = set()
 
@@ -69,6 +73,9 @@ class Input(HorizontalSymbol, object):
         ''' Create the INPUT symbol '''
         ast = ast or ogAST.Input()
         self.branch_entrypoint = None
+        if not ast.pos_y and parent:
+            # Make sure the item is placed below its parent
+            ast.pos_y = parent.y() + parent.boundingRect().height() + 10
         super(Input, self).__init__(parent, text=ast.inputString,
                 x=ast.pos_x, y=ast.pos_y, hyperlink=ast.hyperlink)
         self.set_shape(ast.width, ast.height)
@@ -86,7 +93,6 @@ class Input(HorizontalSymbol, object):
         _, syntax_errors, ___, ____, _____ = self.parser.parseSingleElement(
                 self.common_name, self.pr())
         try:
-            # LOG.error('\n'.join(syntax_errors))
             self.scene().raise_syntax_errors(syntax_errors)
         except AttributeError:
             LOG.debug('raise_syntax_error raised an exception')
@@ -133,8 +139,45 @@ class Input(HorizontalSymbol, object):
         return '\n'.join(result)
 
 
+class Connect(Input):
+    ''' Connect point below a nested state '''
+    common_name = 'connect_part'
+    auto_expand = False
+    resizeable = False
+    # Symbol must not use antialiasing, otherwise the middle line is too thick
+    _antialiasing = False
+    def set_shape(self, width, height):
+        ''' Compute the polygon to fit in width, height '''
+        self.setPen(QPen(Qt.blue))
+        self.textbox_alignment = Qt.AlignLeft | Qt.AlignTop
+        path = QPainterPath()
+        path.moveTo(0, 0)
+        path.lineTo(0, height)
+        #path.moveTo(0, height / 2)
+        #path.lineTo(width, height / 2)
+        self.setPath(path)
+        super(Input, self).set_shape(width, height)
+
+    def resize_item(self, rect):
+        ''' Symbol cannot be resized '''
+        return
+
+    def pr(self):
+        ''' Return the PR notation of the single CONNECT symbol '''
+        comment = repr(self.comment) if self.comment else ';'
+        pos = self.scenePos()
+        pr_string =('/* CIF CONNECT ({x}, {y}), ({w}, {h}) */\n'
+                '{hlink}'
+                'CONNECT {i}{comment}'.format(
+                    hlink=repr(self.text), i=str(self.text),
+                    x=int(pos.x()), y=int(pos.y()),
+                    w=int(self.boundingRect().width()),
+                    h=int(self.boundingRect().height()), comment=comment))
+        return pr_string
+
+
 # pylint: disable=R0904
-class Output(VerticalSymbol, object):
+class Output(VerticalSymbol):
     ''' SDL OUTPUT Symbol '''
     _unique_followers = ['Comment']
     _insertable_followers = [
@@ -184,7 +227,7 @@ class Output(VerticalSymbol, object):
 
 
 # pylint: disable=R0904
-class Decision(VerticalSymbol, object):
+class Decision(VerticalSymbol):
     ''' SDL DECISION Symbol '''
     _unique_followers = ['Comment']
     _insertable_followers = ['DecisionAnswer', 'Task', 'ProcedureCall', 'Output',
@@ -220,7 +263,7 @@ class Decision(VerticalSymbol, object):
         for branch in self.branches():
             if not branch.last_branch_item.terminal_symbol:
                 return False
-        else:        
+        else:
             return True
 
     def branches(self):
@@ -285,8 +328,11 @@ class Decision(VerticalSymbol, object):
             try:
                 # Disconnect the comment of the last item so that
                 # it's size is not taken into account
-                last.comment.setParentItem(None)
-                last.comment.connection.setParentItem(None)
+                pass
+                # Bad idea (MP 05/05/14) Due to Pyside bugs, better avoid
+                # the setParentItem(None) calls
+                #last.comment.setParentItem(None)
+                #last.comment.connection.setParentItem(None)
             except AttributeError:
                 pass
             branch_len = branch.y() + (
@@ -351,7 +397,7 @@ class Decision(VerticalSymbol, object):
 
 
 # pylint: disable=R0904
-class DecisionAnswer(HorizontalSymbol, object):
+class DecisionAnswer(HorizontalSymbol):
     ''' If Decision is a "switch", DecisionAnswer is a "case" '''
     _insertable_followers = ['DecisionAnswer', 'Task', 'ProcedureCall',
                         'Output', 'Decision', 'Label']
@@ -395,8 +441,8 @@ class DecisionAnswer(HorizontalSymbol, object):
         item_parent = (parent if not isinstance(parent, DecisionAnswer)
                        else parent.parentItem())
         super(DecisionAnswer, self).insert_symbol(item_parent, x, y)
-        self.last_branch_item.connectionBelow = Connection(
-                self.last_branch_item, item_parent, connectionPoint=True)
+        self.last_branch_item.connectionBelow = \
+                JoinConnection(self.last_branch_item, item_parent)
 
     def boundingRect(self):
         return QRectF(0, 0, self.width, self.height)
@@ -441,7 +487,7 @@ class DecisionAnswer(HorizontalSymbol, object):
 
 
 # pylint: disable=R0904
-class Join(VerticalSymbol, object):
+class Join(VerticalSymbol):
     ''' JOIN symbol (GOTO) '''
     auto_expand = False
     arrow_head = True
@@ -453,7 +499,7 @@ class Join(VerticalSymbol, object):
 
     def __init__(self, parent=None, ast=None):
         if not ast:
-            ast = ogAST.Terminator(defName='Go')
+            ast = ogAST.Terminator(defName='')
             ast.pos_y = 0
             ast.width = 35
             ast.height = 35
@@ -469,7 +515,7 @@ class Join(VerticalSymbol, object):
         size = min(rect.width(), rect.height())
         rect.setWidth(size)
         rect.setHeight(size)
-        super(Join, self).resize_item(rect) 
+        super(Join, self).resize_item(rect)
 
     def set_shape(self, width, height):
         ''' Define the bouding rectangle of the JOIN symbol '''
@@ -491,14 +537,14 @@ class Join(VerticalSymbol, object):
                     h=int(self.boundingRect().height())))
 
 
-class ProcedureStop(Join, object):
+class ProcedureStop(Join):
     ''' Procedure STOP symbol - very similar to JOIN '''
     # Define reserved keywords for the syntax highlighter
     blackbold = SDL_BLACKBOLD
     redbold = SDL_REDBOLD
     def __init__(self, parent=None, ast=None):
         if not ast:
-            ast = ogAST.Terminator(defName='ret')
+            ast = ogAST.Terminator(defName='')
             ast.pos_y = 0
             ast.width = 35
             ast.height = 35
@@ -517,7 +563,7 @@ class ProcedureStop(Join, object):
         path.moveTo(point1)
         path.lineTo(point2)
         path.moveTo(point3)
-        path.lineTo(point4)        
+        path.lineTo(point4)
         self.setPath(path)
         # call Join superclass, otherwise symbol will take Join shape
         super(Join, self).set_shape(circ, circ)
@@ -525,16 +571,17 @@ class ProcedureStop(Join, object):
     def __repr__(self):
         ''' Return the PR string for this symbol '''
         pos = self.scenePos()
-        return('/* CIF STOP ({x}, {y}), ({w}, {h}) */\n'
+        return('/* CIF RETURN ({x}, {y}), ({w}, {h}) */\n'
                 '{hlink}'
-                'STOP;'.format(hlink=repr(self.text),
+                'RETURN {val};'.format(
+                    val=str(self.text), hlink=repr(self.text),
                     x=int(pos.x()), y=int(pos.y()),
                     w=int(self.boundingRect().width()),
                     h=int(self.boundingRect().height())))
 
 
 # pylint: disable=R0904
-class Label(VerticalSymbol, object):
+class Label(VerticalSymbol):
     ''' LABEL symbol '''
     _insertable_followers = [
             'Task', 'ProcedureCall', 'Output', 'Decision', 'Label']
@@ -618,7 +665,7 @@ class Label(VerticalSymbol, object):
 
 
 # pylint: disable=R0904
-class Task(VerticalSymbol, object):
+class Task(VerticalSymbol):
     ''' TASK symbol '''
     _unique_followers = ['Comment']
     _insertable_followers = [
@@ -667,7 +714,7 @@ class Task(VerticalSymbol, object):
 
 
 # pylint: disable=R0904
-class ProcedureCall(VerticalSymbol, object):
+class ProcedureCall(VerticalSymbol):
     ''' PROCEDURE CALL symbol '''
     _unique_followers = ['Comment']
     _insertable_followers = [
@@ -681,7 +728,7 @@ class ProcedureCall(VerticalSymbol, object):
     completion_list = {'set_timer', 'reset_timer', 'write', 'writeln'}
 
     def __init__(self, parent=None, ast=None):
-        ast = ast or ogAST.Output(defName='callMe')
+        ast = ast or ogAST.Output(defName='')
         super(ProcedureCall, self).__init__(parent,
                 text=ast.inputString, x=ast.pos_x, y=ast.pos_y,
                 hyperlink=ast.hyperlink)
@@ -717,7 +764,7 @@ class ProcedureCall(VerticalSymbol, object):
 
 
 # pylint: disable=R0904
-class TextSymbol(HorizontalSymbol, object):
+class TextSymbol(HorizontalSymbol):
     ''' Text symbol - used to declare variables, etc. '''
     common_name = 'text_area'
     needs_parent = False
@@ -792,10 +839,10 @@ class ASN1Viewer(TextSymbol):
 
 
 # pylint: disable=R0904
-class State(VerticalSymbol, object):
+class State(VerticalSymbol):
     ''' SDL STATE Symbol '''
     _unique_followers = ['Comment']
-    _insertable_followers = ['Input']
+    _insertable_followers = ['Input', 'Connect']
     arrow_head = True
     common_name = 'terminator_statement'
     needs_parent = False
@@ -806,6 +853,7 @@ class State(VerticalSymbol, object):
 
     def __init__(self, parent=None, ast=None):
         ast = ast or ogAST.State()
+        ast.inputString = getattr(ast, 'via', None) or ast.inputString
         # Note: ast coordinates are in scene coordinates
         super(State, self).__init__(parent=parent,
                 text=ast.inputString, x=ast.pos_x, y=ast.pos_y,
@@ -816,8 +864,8 @@ class State(VerticalSymbol, object):
         if parent:
             try:
                 # Map AST scene coordinates to get actual position
-                self.setPos(self.pos() 
-                        + self.mapFromScene(ast.pos_x, ast.pos_y))
+                self.setPos(self.pos()
+                            + self.mapFromScene(ast.pos_x, ast.pos_y))
             except TypeError:
                 self.update_position()
         else:
@@ -827,6 +875,32 @@ class State(VerticalSymbol, object):
         if ast.comment:
             Comment(parent=self, ast=ast.comment)
 
+    @property
+    def allow_nesting(self):
+        ''' Redefinition - must be checked according to context '''
+        result = not any(elem in str(self).lower().strip()
+                       for elem in ('-', ',', '*', ' '))
+        return result
+
+    @property
+    def nested_scene(self):
+        ''' Redefined - nested scene per state must be unique '''
+        return self._nested_scene
+
+    def double_click(self):
+        ''' Catch a double click - Set nested scene '''
+        for each, value in self.scene().composite_states.viewitems():
+            if str(self).lower() == str(each):
+                self.nested_scene = value
+                break
+        else:
+            self.nested_scene = None
+
+    @nested_scene.setter
+    def nested_scene(self, value):
+        ''' Set the value of the nested scene '''
+        self._nested_scene = value
+
     def update_completion_list(self):
         ''' When text was entered, update state completion list '''
         # Get AST for the symbol
@@ -850,6 +924,12 @@ class State(VerticalSymbol, object):
         ''' Compute the polygon to fit in width, height '''
         path = QPainterPath()
         path.addRoundedRect(0, 0, width, height, height / 4, height)
+
+        if self.nested_scene and self.is_composite():
+            # Distinguish composite states with dash line
+            self.setPen(QPen(Qt.DashLine))
+        else:
+            self.setPen(QPen(Qt.SolidLine))
         self.setPath(path)
         super(State, self).set_shape(width, height)
 
@@ -865,6 +945,27 @@ class State(VerticalSymbol, object):
                 'state', self.parse_gr(recursive=True))
             return ast, terminators
 
+
+    def parse_composite_state(self):
+        ''' Return PR string corresponding to the nested part of the state '''
+        entry_points, exit_points = [], []
+        result = ['STATE {};'.format(str(self)),
+                  'SUBSTRUCTURE']
+        for each in self.nested_scene.start:
+            if str(each):
+                entry_points.append(str(each))
+        for each in self.nested_scene.returns:
+            if str(each) != 'no_name':
+                exit_points.append(str(each))
+        if entry_points:
+            result.append('in ({});'.format(','.join(entry_points)))
+        if exit_points:
+            result.append('out ({});'.format(','.join(exit_points)))
+        result.extend(self.nested_scene.get_pr_string())
+        result.append('ENDSUBSTRUCTURE;')
+        return '\n'.join(result)
+
+
     def parse_gr(self, recursive=True):
         ''' Parse state '''
         comment = repr(self.comment) if self.comment else ';'
@@ -905,19 +1006,106 @@ class State(VerticalSymbol, object):
         return '\n'.join(result)
 
 
-class Procedure(HorizontalSymbol, object):
-    ''' Procedure declaration symbol '''
+class Process(HorizontalSymbol):
+    ''' Process symbol '''
     _unique_followers = ['Comment']
+    _allow_nesting = True
+    common_name = 'process_definition'
+    needs_parent = False
+    # Define reserved keywords for the syntax highlighter
+    blackbold = SDL_BLACKBOLD
+    redbold = SDL_REDBOLD
+    completion_list = set()
+    is_singleton = True
+
+    def __init__(self, ast=None, subscene=None):
+        ast = ast or ogAST.Process()
+        super(Process, self).__init__(parent=None,
+                                      text=ast.processName,
+                                      x=ast.pos_x,
+                                      y=ast.pos_y,
+                                      hyperlink=ast.hyperlink)
+        self.set_shape(ast.width, ast.height)
+        self.setBrush(QBrush(QColor(255, 255, 202)))
+        self.parser = ogParser
+        if ast.comment:
+            Comment(parent=self, ast=ast.comment)
+        self.nested_scene = subscene
+
+    def set_shape(self, width, height):
+        ''' Compute the polygon to fit in width, height '''
+        path = QPainterPath()
+        path.moveTo(7, 0)
+        path.lineTo(0, 7)
+        path.lineTo(0, height - 7)
+        path.lineTo(7, height)
+        path.lineTo(width - 7, height)
+        path.lineTo(width, height - 7)
+        path.lineTo(width, 7)
+        path.lineTo(width - 7, 0)
+        path.lineTo(7, 0)
+        self.setPath(path)
+        super(Process, self).set_shape(width, height)
+
+    def recursive_parser(self):
+        ''' Return the full content of the process as PR string '''
+        pr_data = deque()
+        scene = self.nested_scene
+        if not scene:
+            return []
+
+        for item in chain(scene.texts, scene.procs, scene.start):
+            pr_data.append(repr(item))
+        for item in scene.floating_labels:
+            pr_data.append(item.parse_gr())
+        composite = set(scene.composite_states.keys())
+        for item in scene.states:
+            if item.is_composite():
+                try:
+                    composite.remove(str(item).lower())
+                    pr_data.appendleft(item.parse_composite_state())
+                except KeyError:
+                    pass
+            pr_data.append(item.parse_gr())
+        return list(pr_data)
+
+    def parse_gr(self, recursive=True):
+        ''' Generate PR for a process '''
+        comment = repr(self.comment) if self.comment else ';'
+        pos = self.scenePos()
+        result =['/* CIF PROCESS ({x}, {y}), ({w}, {h}) */\n'
+                 '{hlink}'
+                 'PROCESS {process}{comment}'.format(
+                        process=str(self),
+                        hlink=repr(self.text),
+                        x=int(pos.x()), y=int(pos.y()),
+                        w=int(self.boundingRect().width()),
+                        h=int(self.boundingRect().height()), comment=comment)]
+        if recursive:
+            result.append('\n'.join(self.recursive_parser()))
+        result.append('ENDPROCESS {};'.format(str(self)))
+        return '\n'.join(result)
+
+    def __repr__(self):
+        ''' Return the complete process content '''
+        return self.parse_gr(recursive=True)
+
+
+class Procedure(Process):
+    ''' Procedure declaration symbol - Very similar to Process '''
+    _unique_followers = ['Comment']
+    _allow_nesting = True
     common_name = 'procedure'
     needs_parent = False
     # Define reserved keywords for the syntax highlighter
     blackbold = SDL_BLACKBOLD
     redbold = SDL_REDBOLD
     completion_list = set()
+    is_singleton = False
 
     def __init__(self, ast=None, subscene=None):
         ast = ast or ogAST.Procedure()
-        super(Procedure, self).__init__(parent=None,
+        super(Process, self).__init__(parent=None,
                 text=ast.inputString,
                 x=ast.pos_x, y=ast.pos_y, hyperlink=ast.hyperlink)
         self.set_shape(ast.width, ast.height)
@@ -952,17 +1140,14 @@ class Procedure(HorizontalSymbol, object):
         pos = self.scenePos()
         result = ['/* CIF PROCEDURE ({x}, {y}), ({w}, {h}) */\n'
                     '{hlink}'
-                    'PROCEDURE {state}{comment}'.format(
-                        state=str(self),
+                    'PROCEDURE {proc}{comment}'.format(
+                        proc=str(self),
                         hlink=repr(self.text),
                         x=int(pos.x()), y=int(pos.y()),
                         w=int(self.boundingRect().width()),
                         h=int(self.boundingRect().height()), comment=comment)]
         if recursive:
-            pr_raw = self.nested_scene.get_pr_string()
-            # Keep only the body - it is not a PROCESS
-            pr_data = pr_raw[1:-1]
-            result.extend(pr_data)
+            result.append('\n'.join(self.recursive_parser()))
         result.append('ENDPROCEDURE;')
         return '\n'.join(result)
 
@@ -971,9 +1156,8 @@ class Procedure(HorizontalSymbol, object):
         return self.parse_gr(recursive=True)
 
 
-
 # pylint: disable=R0904
-class Start(HorizontalSymbol, object):
+class Start(HorizontalSymbol):
     ''' SDL START Symbol '''
     _unique_followers = ['Comment']
     _insertable_followers = [
@@ -986,14 +1170,16 @@ class Start(HorizontalSymbol, object):
     # Define reserved keywords for the syntax highlighter
     blackbold = SDL_BLACKBOLD
     redbold = SDL_REDBOLD
+    has_text_area = False
 
     def __init__(self, ast=None):
         ''' Create the START symbol '''
         ast = ast or ogAST.Start()
         self.terminal_symbol = False
-        super(Start, self).__init__(parent=None, text='',
-                x=ast.pos_x, y=ast.pos_y,
-                hyperlink=ast.hyperlink)
+        super(Start, self).__init__(parent=None,
+                                    text=ast.inputString[slice(0,-6)],
+                                    x=ast.pos_x, y=ast.pos_y,
+                                    hyperlink=ast.hyperlink)
         self.set_shape(ast.width, ast.height)
         self.setBrush(QBrush(QColor(255, 228, 213)))
         # No hyperlink for START symbol because it has no text
@@ -1019,7 +1205,9 @@ class Start(HorizontalSymbol, object):
         comment = repr(self.comment) if self.comment else ';'
         pos = self.scenePos()
         result.append('/* CIF START ({x}, {y}), ({w}, {h}) */\n'
-                'START{comment}'.format(x=int(pos.x()), y=int(pos.y()),
+                'START{via}{comment}'.format(x=int(pos.x()), y=int(pos.y()),
+                    via = (' ' + str(self) + ' ')
+                    if str(self).replace('START', '') else '',
                     w=int(self.boundingRect().width()),
                     h=int(self.boundingRect().height()), comment=comment))
         # Recursively return the complete branch below the start symbol
@@ -1030,7 +1218,7 @@ class Start(HorizontalSymbol, object):
         return '\n'.join(result)
 
 
-class ProcedureStart(Start, object):
+class ProcedureStart(Start):
     ''' Start symbol of a procedure - only shape differs from Start '''
     # Define reserved keywords for the syntax highlighter
     blackbold = SDL_BLACKBOLD
@@ -1046,3 +1234,13 @@ class ProcedureStart(Start, object):
         path.lineTo(max(width / 2, width - height / 2), height)
         self.setPath(path)
         super(Start, self).set_shape(width, height)
+
+
+class StateStart(Start):
+    ''' Composite states can have several named START symbols '''
+    has_text_area = True
+    is_singleton = False
+
+    def __str__(self):
+        ''' Return the state entry point '''
+        return str(self.text)
